/*
 fmuladd CPU code to emulate nVidia's fmuladd CUDA open-source code, on Intel SSE CPU.
 The computation is done in two-ways, using floating-point normal code (that GCC may translate to bad SSE code),
 and native SSE code.
 
 As the original nVidia code, this one send back the sustained Gflop from the current CPU core.
 As it is not multithreaded, Total Gflop of the CPU is the GFlop/core (output of this program) * number of core.
 ie: on a Core2 Duo 2.4Ghz Penryn, this program will output nearly 9Gflops/core, so the CPU is capable of 18Glops.
 
 How-to compile:
 gcc -O3 -o fmuladd-cpu fmuladd-cpu.c
 (Intel SSE _m128 & SSE operators seems not to be correctly supported on VisualStudio nor other proprietary closed-source compiler,
  so I suggest testing using GCC 4.0.1 or later).
 
 This is open-source, feel free in any way to modify it, comment it, distribute it, sell it, use it, exploit it commercially, 
 as long as you cite the author : philippe vigier/iAPX
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include <xmmintrin.h>

#include <time.h>

#define NUM_SMS (16)
#define NUM_THREADS_PER_SM (384)
#define NUM_THREADS_PER_BLOCK (192)
#define NUM_BLOCKS ((NUM_THREADS_PER_SM / NUM_THREADS_PER_BLOCK) * NUM_SMS)
#define NUM_ITERATIONS 10



// FMAD8 = 8 MUL + 8 ADD
#define FMAD8(a, b, c, d, e, f, g, h) \
     a = a * b + b;		\
	 c = c * d + d;		\
	 e = e * f + f;		\
	 g = g * h + h;		\
	 b = a * b + a;		\
	 d = c * d + c;		\
	 f = e * f + e;		\
	 h = h * g + g;
	 


void gflops_x87( float *pValue )
{
  int i;
  
  float		a, b, c, d, e, f, g, h;
  a = *pValue;
  b = 1.01f;
  c = *pValue;
  d = 1.0002f;
  e = *pValue;
  f = 1.003f;
  g = *pValue;
  h = 1.005f;

   for (i = 0; i < NUM_ITERATIONS * 16 * 16; i++)
   {
       FMAD8(a, b, c , d, e, f, g, h);
   }
   
   // To avoid GCC to throw computation ;-)
   *pValue = b + d + f +h;
}




// The SSE version for muladd16 (8 muladd * 4 float = 32 muladd) 
#define FMADDSSE32(a, b, c, d, e, f, g, h) \
     a = a * b + b; \
	 c = c * d + d; \
	 e = e * f + f; \
	 g = g * h + h; \
	 b = a * b + a; \
	 d = c * d + c; \
	 f = e * f + e; \
	 h = g * h + g;


void gflops_sse( float *pValue )
{
  int i;
  
   __m128	a, b, c, d, e, f, g, h;
   a = _mm_set_ps1( (*pValue) );
   b = _mm_set_ps1( 1.01f );
   c = _mm_set_ps1( (*pValue) );
   d = _mm_set_ps1( 1.01f );
   e = _mm_set_ps1( (*pValue) );
   f = _mm_set_ps1( 1.01f );
   g = _mm_set_ps1( (*pValue) );
   h = _mm_set_ps1( 1.01f );
   
   for (i = 0; i < NUM_ITERATIONS * 16 * 4; i++)
   {
       FMADDSSE32(a, b, c, d, e, f, g, h);
   }
   
   // To avoid GCC to throw computations ;-)
   b = a+b+c+d+e+f+g+h;
   *pValue = * (float * ) &b;
}


int
main(int argc, char** argv)
{
  int i;
  clock_t start, end;
  double cpu_time_used_ms;
  float initValue = rand();
  float flops;
  

/********************************** CPU x87 Part *******************************/
   start = clock();
   for(i=0; i<NUM_BLOCKS*NUM_THREADS_PER_BLOCK*10; i++) {
     gflops_x87( &initValue );
   }
   
   // How much time did we spend on that?!?
   end = clock();
   cpu_time_used_ms = ((double) (end - start)) * 1000.0f / CLOCKS_PER_SEC;
   printf("CPU-x87 Time used = %4.3f ms \n", (float) cpu_time_used_ms );

   // And now the Gflop value :-)
   flops = 128 * 2 * 16 * NUM_ITERATIONS * NUM_BLOCKS * NUM_THREADS_PER_BLOCK;
   printf("CPU-x87 : %3.3f Gflops/core\n", (flops / cpu_time_used_ms ) / 1e5 );


/********************************** CPU SSE Part *******************************/
   start = clock();
   for(i=0; i<NUM_BLOCKS*NUM_THREADS_PER_BLOCK*10; i++) {
     gflops_sse( &initValue );
   }
   
   // How much time did we spend on that?!?
   end = clock();
   cpu_time_used_ms = ((double) (end - start)) * 1000.0f / CLOCKS_PER_SEC;
   printf("CPU-SSE Time used = %4.3f ms \n", (float) cpu_time_used_ms );

   // And now the Gflop value :-)
   flops = 128 * 2 * 16 * NUM_ITERATIONS * NUM_BLOCKS * NUM_THREADS_PER_BLOCK;
   printf("CPU-SSE : %3.3f Gflops/core\n", (flops / cpu_time_used_ms ) / 1e5 );
   
   
   // exit, forcing GCC to not discard our computations ;-)
   exit( initValue < 0 ? 1 : 0 );
}

